// =====================================================================================
// 
//       Filename:  Distribution.hpp
// 
//    Description:  The definition of class Distribution
// 
//        Version:  1.0
//        Created:  02/06/2010 12:31:00 PM
//       Revision:  none
//       Compiler:  g++
// 
//         Author:  Jianxing Feng (feeldead), feeldead@gmail.com
//        Company:  THU
// 
// =====================================================================================

#ifndef Distribution_H
#define Distribution_H

#include <gsl/gsl_cdf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <cassert>

#include <iostream>
#include "Utility.hpp"

using namespace std;

class Distribution
{
	public:
		virtual bool IsConstant() = 0;
		// Return the probability of the value in range [low, high)
		virtual double ProbInRange(double low, double high) = 0;
		// P{X <= value}
		virtual double ProbLessThan(double value) = 0;
		virtual double GetValue(gsl_rng* rng = 0) = 0;
		virtual double RangeLow() = 0;
		virtual double RangeHigh() = 0;
		virtual void Write(ostream* p_out) = 0;
		virtual void Read(istream* p_in) = 0;
};

class DistributionConstant : public Distribution
{
	public:
		double mValue;

	public:
		virtual bool IsConstant(){return true;}

		virtual double ProbInRange(double low, double high)
		{
			if (mValue >= low && mValue < high) return 1;
			return 0;
		}
		virtual double ProbLessThan(double value)
		{
			if (mValue < value) return 1;
			return 0;
		}
		virtual double GetValue(gsl_rng* rng = 0){return mValue;}

		virtual double RangeLow(){return mValue;}
		virtual double RangeHigh() {return mValue+1;}

		virtual void Write(ostream* p_out)
		{
			(*p_out) << "0" << endl << mValue<< " ";
		}
		virtual void Read(istream* p_in)
		{
			(*p_in) >> mValue;
		}
};

class DistributionGaussian: public Distribution
{
	private:
		double mMean;
		double mStd;

	public:
		virtual bool IsConstant(){return false;}

		virtual double ProbInRange(double low, double high)
		{
			double prob_in_range = gsl_cdf_gaussian_P((high - mMean) / mStd, 1) -
								   gsl_cdf_gaussian_P((low - mMean) / mStd, 1);
			return prob_in_range;
		}
		virtual double GetValue(gsl_rng* rng = 0)
		{
			// Randomly select a value following N(span_mean, span_std)
			if (rng)
			{
				return (gsl_ran_gaussian(rng, 1) * mStd + mMean);
			}
			else
				return mMean;
		}
		virtual double ProbLessThan(double value)
		{
			return gsl_cdf_gaussian_P((value - mMean) / mStd, 1);
		}

		// Allows three standard deviations. About 99.7%
		virtual double RangeLow(){return mMean - 3 * mStd;}
		virtual double RangeHigh() {return mMean + 3 * mStd + 1;}

		virtual void Write(ostream* p_out)
		{
			(*p_out) << "1" << endl;
			(*p_out) << mMean << " ";
			(*p_out) << mStd << " ";
		}
		virtual void Read(istream* p_in)
		{
			(*p_in) >> mMean;
			(*p_in) >> mStd;
		}
};

class DistributionCustomize: public Distribution
{
	private:
		int mValueCnt;
		vector<unsigned> mValues;
		vector<double> mProbabilities;
		vector<double> mPartSum;

		double mMean;

	public:
		virtual bool IsConstant(){return false;}

		virtual double ProbInRange(double low, double high)
		{
			int start_idx = UtilityTemp<unsigned>::BinarySearch(mValues, low);
			int end_idx = UtilityTemp<unsigned>::BinarySearch(mValues, high);
			// sum over mProbabilities[start_idx,...,end_idx-1]

			if (start_idx >= end_idx) return 0;
			return mPartSum[end_idx] - mPartSum[start_idx];
		}
		virtual double GetValue(gsl_rng* rng = 0)
		{
			// Randomly select a value following N(span_mean, span_std)
			if (rng)
			{
				// Randomly select a start position of the read
				double rand = gsl_rng_uniform(rng);
				int idx = UtilityTemp<double>::BinarySearch(mPartSum, rand);
				if (idx > 0) --idx;
				return mValues[idx];
			}
			else
				return mMean;
		}
		virtual double ProbLessThan(double value)
		{
			int idx = UtilityTemp<unsigned>::BinarySearch(mValues, value);
			return mPartSum[idx];
		}

		virtual double RangeLow(){return mValues[0];}
		virtual double RangeHigh() {return (mValues[mValues.size()-1] + 1);}

		virtual void Write(ostream* p_out)
		{
			(*p_out) << "2" << endl;
			(*p_out) << mValues.size() << endl;
			for (unsigned i = 0; i < mValues.size(); ++i)
			{
				(*p_out) << mValues[i] << "\t" << mProbabilities[i];
				if (i < mValues.size() - 1) (*p_out) << endl;
			}
		}
		virtual void Read(istream* p_in)
		{
			int value_cnt;
			(*p_in) >> value_cnt;
			mValues.resize(value_cnt);
			mProbabilities.resize(value_cnt);

			for (int i = 0; i < value_cnt; ++i)
			{
				(*p_in) >> mValues[i];
				(*p_in) >> mProbabilities[i];
			}

			vector<int> sortedIdx;
			UtilityTempComp<unsigned, less<unsigned> >::Sort(mValues, sortedIdx);
			UtilityTemp<double>::SortByIndex(mProbabilities, sortedIdx);

			// mPartSum[i] = sum mProbabilities[0,...,i-1]
			mPartSum.resize(value_cnt+1);
			mPartSum[0] = 0;
			for (int i = 0; i < value_cnt; ++i)
				mPartSum[i+1] = mPartSum[i] + mProbabilities[i];

			cout << "Custom distribution: " << value_cnt << " value-probability pairs have been loaded" << endl;
			cout << "The sum of all the probabilities of all the lengths is " << mPartSum[value_cnt] << endl;

			if (1 != mPartSum[value_cnt]) 
			{
				// Scale the probability
				double scale = 1 / mPartSum[value_cnt];
				for (unsigned i = 0; i < value_cnt; ++i)
				{
					mPartSum[value_cnt+1] *= scale;
					mProbabilities[value_cnt] *= scale;
				}
				cout << "Probabilities are scaled" << endl;
			}

			mMean = 0;
			for (int i = 0; i < value_cnt; ++i)
				mMean += mValues[i] * mProbabilities[i];
		}
};

#endif
